
import time
import subprocess
from argparse import ArgumentParser
from typing import Callable
from rkd.api.contract import TaskInterface, ExecutionContext


class WaitForDatabaseTask(TaskInterface):
    """ Waits for the database to be ready """

    def get_name(self) -> str:
        return ':wait-for'

    def get_group_name(self) -> str:
        return ':db'

    def execute(self, context: ExecutionContext) -> bool:
        return self.check_if_instance_is_alive(
            db_type=context.args['type'],
            db_name=context.args['db_name'],
            timeout=int(context.args['timeout']),
            user=context.args['username'],
            password=context.args['password'],
            port=int(context.args['port']),
            host=context.args['host']
        )

    def _get_checker_command(self, db_type: str, db_name: str, user: str,
                             password: str, host: str, port: int) -> Callable:
        """
        Detect which checking command fits best
        :return:
        """

        if db_type == 'mysql':
            if self.is_mysql_tool_available() and user:
                self._io.debug(' >> MySQL tool checker was selected')
                return self.get_wait_command_for_mysql(host=host, port=port, user=user, password=password)

            self._io.debug(' >> NC tool was selected')
            return self.get_wait_command_for_mysql_nc(host=host, port=port)

        if db_type == 'postgres':
            if self.is_postgres_toolkit_available():
                self._io.debug(' >> PostgreSQL toolkit is available and will be used')
                return self.get_wait_command_for_pg_is_ready(db_name=db_name, user=user, host=host, port=port)

            self._io.debug(' >> NC tool was selected for PostgreSQL')
            return self.get_wait_command_for_postgres_nc(host=host, port=port)

        raise Exception('Unsupported database type. Only mysql and postgres are supported.')

    def check_if_instance_is_alive(self, db_type: str, db_name: str, user: str,
                                   password: str, host: str, port: int, timeout: int) -> bool:

        checker = self._get_checker_command(db_type=db_type, db_name=db_name, user=user, password=password,
                                            host=host, port=port)
        time_left = timeout

        while time_left != 0:
            if checker():
                self._io.success_msg(' >> The DB seems to be online')
                return True

            time.sleep(1)
            time_left -= 1

        self._io.error_msg(' >> Error: The DB is still down')
        return False

    def get_wait_command_for_mysql(self, host: str, port: int, user: str, password: str) -> Callable:
        return lambda: self.is_command_of_success_code(
            'mysql ' +
            '-h "' + host + '" ' +
            '-P "' + str(port) + '" ' +
            '-u"' + user + '" ' +
            ('-p"' + password + '" ' if password else ' ') +
            '-e "SELECT 1;"'
        )

    def get_wait_command_for_mysql_nc(self, host: str, port: int) -> Callable:
        return lambda: self.is_command_of_success_code(
            'echo "ttttt\n\n" | nc -w 1 "%s" "%i" | grep "mysql_native"' % (
                host, port
            )
        )

    def get_wait_command_for_postgres_nc(self, host: str, port: int) -> Callable:
        return lambda: self.is_command_of_success_code(
            'nc -z %s %i' % (host, port)
        )

    def get_wait_command_for_pg_is_ready(self, db_name: str, user: str, host: str, port: int) -> Callable:
        opts = ''

        if db_name:
            opts += ' --dbname=%s ' % db_name

        if user:
            opts += ' --username=%s ' % user

        return lambda: self.is_command_of_success_code(
            'pg_isready %s --host=%s --port=%i --timeout=1' % (opts, host, port)
        )

    def is_mysql_tool_available(self) -> bool:
        return self.is_command_of_success_code('command -v mysql')

    def is_postgres_toolkit_available(self) -> bool:
        return self.is_command_of_success_code('command -v pg_isready')

    @staticmethod
    def is_command_of_success_code(cmd: str) -> bool:
        try:
            subprocess.check_output(cmd, shell=True)
            return True

        except subprocess.CalledProcessError as e:
            print(e.output.decode('utf-8'))
            return False

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--host', '-a',
                            help='IP address or UNIX socket',
                            required=True)
        parser.add_argument('--port', '-p',
                            help='Port',
                            required=True)
        parser.add_argument('--type', '-t',
                            help='Database type. Possible values: postgres, mysql',
                            required=True)
        parser.add_argument('--username', '-u',
                            help='Username (optional, when not specified, then only port will be checked)',
                            required=False)
        parser.add_argument('--db-name', '-n',
                            help='Database name (optional)',
                            default='')
        parser.add_argument('--password', '-P',
                            help='Password (optional)',
                            default='')
        parser.add_argument('--timeout', '-T',
                            help='Timeout in seconds (optional, defaults to 15)',
                            default=15)




import json
from argparse import ArgumentParser
from rkd.api.contract import TaskInterface, ExecutionContext


class EnvToJsonTask(TaskInterface):
    """ Dump environment variables into JSON """

    def get_name(self) -> str:
        return ':env-to-json'

    def get_group_name(self) -> str:
        return ':utils'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--parse-json', '-p',
                            help='Parse JSON values of all variables which have JSON as value',
                            action='store_true')

    def execute(self, context: ExecutionContext) -> bool:
        env = dict(context.env)
        to_serialize = {}

        for name, value in env.items():
            if context.args['parse_json']:
                try:
                    value = json.loads(value)
                except json.JSONDecodeError:
                    pass

            to_serialize[name] = value

        self._io.out(json.dumps(to_serialize, indent=2, sort_keys=True))
        return True




import re
from argparse import ArgumentParser
from abc import ABC
from typing import Callable
from subprocess import CalledProcessError
from rkd.api.contract import TaskInterface, ExecutionContext
from rkd.api.syntax import TaskDeclaration


class DockerBaseTask(TaskInterface, ABC):
    def calculate_images(self, image: str, latest_per_version: bool, global_latest: bool, allowed_meta_list: str):
        """ Calculate tags propagation """

        allowed_meta = allowed_meta_list.replace(' ', '').split(',')
        tag = image.split(':')[-1]

        # output
        output_tags = [image]

        matches = re.match('([0-9.]+)(-([A-Za-z]+))?([0-9]+)?', tag, re.IGNORECASE)
        meta_type = matches.group(3)

        if not matches:
            self._io.warn('No release version found')
            return output_tags

        if meta_type and meta_type not in allowed_meta:
            self.io().warn('Version meta part is not allowed, not calculating propagation')
            return output_tags

        original_tag = matches.group(0)
        base_version = matches.group(1)
        meta = meta_type
        meta_number = matches.group(4)

        # :latest
        if global_latest:
            output_tags.append(image.replace(original_tag, 'latest'))

        # case 1: 1.0.0-RC1 -> 1.0.0-latest-RC
        if meta and meta_number:
            output_tags.append(image.replace(original_tag, base_version + '-latest%s' % meta))

            if latest_per_version:
                output_tags = self._generate_for_each_version(
                    image, original_tag, output_tags,
                    lambda version: original_tag.replace(base_version + meta + meta_number, version + '-latest%s' % meta)
                )
        elif meta and not meta_number:
            output_tags.append(image.replace(original_tag, base_version + '-latest%s' % meta))

            if latest_per_version:
                output_tags = self._generate_for_each_version(
                    image, original_tag, output_tags,
                    lambda version: original_tag.replace(base_version + meta, version + '-latest%s' % meta)
                )
        # release
        elif not meta:
            output_tags = self._generate_for_each_version(
                image, original_tag, output_tags,
                lambda version: original_tag.replace(base_version, version)
            )

        return output_tags

    @staticmethod
    def _generate_for_each_version(image: str, original_tag: str, output_tags: list, callback: Callable) -> list:
        parts = original_tag.split('.')

        for part_num in range(0, len(parts)):
            version = ".".join(parts[0:part_num])

            if not version:
                continue

            output_tags.append(
                image.replace(
                    original_tag,
                    callback(version)
                )
            )

        return output_tags

    def _print_images(self, images: list, action: str):
        for image in images:
            self._io.info(' -> Going to %s image "%s"' % (action, image))

    def get_group_name(self) -> str:
        return ':docker'

    def configure_argparse(self, parser: ArgumentParser):
        parser.add_argument('--image', '-i', help='Image name', required=True)
        parser.add_argument('--without-latest', '-wl', help='Do not tag latest per version', action='store_true')
        parser.add_argument('--without-global-latest', '-wgl', help='Do not tag :latest', action='store_true')
        parser.add_argument('--propagate', '-p', help='Propagate tags? eg. 1.0.0 -> 1.0 -> 1 -> latest', action='store_true')
        parser.add_argument('--allowed-meta', '-m', help='Allowed meta part eg. rc, alpha, beta',
                            default='rc,alpha,stable,dev,prod,test,beta,build,b')


class TagImageTask(DockerBaseTask):
    """Re-tag images to propagate version tags in docker-like format eg. 1.0.1 -> 1.0 -> 1 -> latest

    Examples:
        1.0.0 -> 1.0 -> 1 -> latest
        1.0.0-RC1 -> 1.0.0-latest-rc
    """

    def get_name(self) -> str:
        return ':tag'

    def execute(self, context: ExecutionContext) -> bool:
        original_image = context.args['image']

        if context.args['propagate']:
            images = self.calculate_images(
                image=original_image,
                latest_per_version=not context.args['without_latest'],
                global_latest=not context.args['without_global_latest'],
                allowed_meta_list=context.args['allowed_meta']
            )
        else:
            images = [original_image]

        self._print_images(images, 'tag')

        for image in images:
            try:
                self.exec('docker tag %s %s' % (original_image, image))
            except CalledProcessError as e:
                print(e)
                return False

        return True


class PushTask(DockerBaseTask):
    """Pushes all re-tagged images
    """

    def get_name(self) -> str:
        return ':push'

    def execute(self, context: ExecutionContext) -> bool:
        original_image = context.args['image']
        images = []

        if context.args['propagate']:
            images += self.calculate_images(
                image=original_image,
                latest_per_version=not context.args['without_latest'],
                global_latest=not context.args['without_global_latest'],
                allowed_meta_list=context.args['allowed_meta']
            )
        else:
            images = [original_image]

        self._print_images(images, 'push')

        for image in images:
            try:
                self.exec('docker push %s' % image)
            except CalledProcessError as e:
                print(e)
                return False

        return True


def imports():
    return [
        TaskDeclaration(TagImageTask()),
        TaskDeclaration(PushTask())
    ]
